"""Stream type classes for tap-neon."""

from __future__ import annotations

import typing as t

from tap_neon.client import NeonStream

if t.TYPE_CHECKING:
    from singer_sdk.helpers.types import Context, Record


__all__ = [
    "Projects",
    "Operations",
    "Branches",
    "Databases",
    "Roles",
    "Endpoints",
]


class Projects(NeonStream):
    """Projects stream."""

    name = "projects"
    path = "/projects"
    primary_keys = ("id",)
    replication_key = None
    swagger_ref = "ProjectListItem"
    records_jsonpath = "$.projects[*]"

    def get_child_context(
        self,
        record: Record,
        context: Context | None,  # noqa: ARG002
    ) -> dict[str, t.Any]:
        """Return the child context for this record.

        Args:
            record: The record to get the child context for.
            context: The parent context.

        Returns:
            The child context.
        """
        return {"project_id": record["id"]}

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        """Initialize the projects stream."""
        super().__init__(*args, **kwargs)
        self._schema["properties"]["default_endpoint_settings"]["properties"][
            "autoscaling_limit_min_cu"
        ]["minimum"] = 0
        self._schema["properties"]["default_endpoint_settings"]["properties"][
            "autoscaling_limit_max_cu"
        ]["minimum"] = 0


class Operations(NeonStream):
    """Operations stream."""

    name = "operations"
    path = "/projects/{project_id}/operations"
    primary_keys = ("id",)
    replication_key = None
    swagger_ref = "Operation"
    records_jsonpath = "$.operations[*]"
    next_page_token_jsonpath = "$.pagination.cursor"  # noqa: S105
    parent_stream_type = Projects


class Branches(NeonStream):
    """Branches stream."""

    name = "branches"
    path = "/projects/{project_id}/branches"
    primary_keys = ("id",)
    replication_key = None
    swagger_ref = "Branch"
    records_jsonpath = "$.branches[*]"
    parent_stream_type = Projects

    def get_child_context(
        self,
        record: Record,
        context: Context | None,  # noqa: ARG002
    ) -> dict[str, t.Any]:
        """Add branch_id to context.

        Args:
            record: A record dict.
            context: A context dict.

        Returns:
            The updated context dict.
        """
        return {
            "project_id": record["project_id"],
            "branch_id": record["id"],
        }


class Databases(NeonStream):
    """Databases stream."""

    name = "databases"
    path = "/projects/{project_id}/branches/{branch_id}/databases"
    primary_keys = ("id",)
    replication_key = None
    swagger_ref = "Database"
    records_jsonpath = "$.databases[*]"
    parent_stream_type = Branches


class Roles(NeonStream):
    """Roles stream."""

    name = "roles"
    path = "/projects/{project_id}/branches/{branch_id}/roles"
    primary_keys = ("name",)
    replication_key = None
    swagger_ref = "Role"
    records_jsonpath = "$.roles[*]"
    parent_stream_type = Branches


class Endpoints(NeonStream):
    """Endpoints stream."""

    name = "endpoints"
    path = "/projects/{project_id}/endpoints"
    primary_keys = ("id",)
    replication_key = None
    swagger_ref = "Endpoint"
    records_jsonpath = "$.endpoints[*]"
    parent_stream_type = Projects

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        """Initialize the endpoints stream."""
        super().__init__(*args, **kwargs)
        self._schema["properties"]["pooler_mode"]["enum"].append("")



"""Neon Serverless Postgres tap class."""

from __future__ import annotations

import typing as t
from copy import deepcopy

import requests
from singer_sdk import Stream, Tap
from singer_sdk import typing as th
from singer_sdk._singerlib import resolve_schema_references

from tap_neon import streams

if t.TYPE_CHECKING:
    from singer_sdk.streams import RESTStream

    from tap_neon.client import NeonStream

OPENAPI_URL = "https://dfv3qgd2ykmrx.cloudfront.net/api_spec/release/v2.json"
STREAMS: list[type[NeonStream]] = [
    streams.Operations,
    streams.Projects,
    streams.Branches,
    streams.Endpoints,
    streams.Roles,
    streams.Databases,
]


def not_required_null(schema: dict[str, t.Any]) -> dict[str, t.Any]:
    """Add 'null' to the type of all properties that are not required."""
    new_schema = deepcopy(schema)
    schema_type: str | list[str] = new_schema.get("type")  # type: ignore[assignment]
    if "object" not in schema_type:
        errmsg = "Schema type must be of 'object' type"
        raise ValueError(errmsg)

    required = new_schema.get("required", [])

    for prop in new_schema.get("properties", {}).values():
        prop_type: str | list[str] = prop.pop("type", [])
        if prop not in required and "null" not in prop_type:
            prop["type"] = (
                [prop_type, "null"]
                if isinstance(prop_type, str)
                else [*prop_type, "null"]
            )

        if "object" in prop_type:
            prop.update(not_required_null(prop))

    return new_schema


class TapNeon(Tap):
    """Singer tap for Neon Serverless Postgres."""

    name = "tap-neon"

    config_jsonschema = th.PropertiesList(
        th.Property(
            "api_key",
            th.StringType,
            required=True,
            description="API Key for Neon Serverless Postgres",
        ),
        th.Property(
            "start_date",
            th.DateTimeType,
            description="Earliest datetime to get data from",
        ),
    ).to_dict()

    def get_openapi_schema(self) -> dict[t.Any, t.Any]:
        """Retrieve Swagger/OpenAPI schema for this API.

        Returns:
            OpenAPI schema.
        """
        return requests.get(OPENAPI_URL, timeout=5).json()  # type: ignore[no-any-return]

    def discover_streams(self) -> list[Stream]:
        """Return a list of discovered streams.

        Returns:
            A list of Neon Serverless Postgres streams.
        """
        streams: list[RESTStream[str]] = []
        openapi_schema = self.get_openapi_schema()

        for stream_type in STREAMS:
            schema = {
                "$ref": f"#/components/schemas/{stream_type.swagger_ref}",
                "components": openapi_schema["components"],
            }
            resolved_schema = not_required_null(resolve_schema_references(schema))
            streams.append(stream_type(tap=self, schema=resolved_schema))

        return sorted(streams, key=lambda x: x.name)



"""Tests standard tap features using the built-in SDK tests library."""

from __future__ import annotations

from singer_sdk.testing import get_tap_test_class

from tap_neon.tap import TapNeon

TestTapNeon = get_tap_test_class(TapNeon, config={})



"""REST client handling, including NeonStream base class."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from singer_sdk import RESTStream
from singer_sdk.authenticators import BearerTokenAuthenticator

if TYPE_CHECKING:
    from singer_sdk.helpers.types import Context


class NeonStream(RESTStream[str]):
    """Neon Serverless Postgres stream class."""

    url_base = "https://console.neon.tech/api/v2"
    next_page_token_jsonpath = "$.next_page"  # noqa: S105

    swagger_ref: str

    @property
    def authenticator(self) -> BearerTokenAuthenticator:
        """Get an authenticator object.

        Returns:
            The authenticator instance for this REST stream.
        """
        token: str = self.config["api_key"]
        return BearerTokenAuthenticator.create_for_stream(
            self,
            token=token,
        )

    @property
    def http_headers(self) -> dict[str, str]:
        """Return the http headers needed.

        Returns:
            A dictionary of HTTP headers.
        """
        return {
            "User-Agent": f"{self.tap_name}/{self._tap.plugin_version}",
            "Content-Type": "application/json",
        }

    def get_url_params(
        self,
        context: Context | None,  # noqa: ARG002
        next_page_token: str | None,
    ) -> dict[str, Any]:
        """Get URL query parameters.

        Args:
            context: Stream sync context.
            next_page_token: Next offset.

        Returns:
            Mapping of URL query parameters.
        """
        params: dict[str, Any] = {}

        if self.next_page_token_jsonpath:
            params["limit"] = 100
            if next_page_token:
                params["cursor"] = next_page_token
        return params
